# -*- coding: utf-8 -*-

"""PyKEEN is a package for training and evaluating knowledge graph embedding.

Installation
------------
Get the Latest
~~~~~~~~~~~~~~~
Download the most recent code from ... with:

For Developers
~~~~~~~~~~~~~~
Clone the repository from `GitHub <https://github.com/SmartDataAnalytics/pykeen.git>'
"""

from .constants import get_version  # noqa: F401
from .run import Results, run  # noqa: F401



# -*- coding: utf-8 -*-

"""Implementation of TransD."""

from dataclasses import dataclass
from typing import Dict

import numpy as np
import torch
import torch.autograd
from torch import nn

from pykeen.constants import RELATION_EMBEDDING_DIM, SCORING_FUNCTION_NORM, TRANS_D_NAME
from pykeen.kge_models.base import BaseModule, slice_triples

__all__ = [
    'TransD',
    'TransDConfig',
]


@dataclass
class TransDConfig:
    relation_embedding_dim: int
    scoring_function_norm: str

    @classmethod
    def from_dict(cls, config: Dict) -> 'TransDConfig':
        """Generate an instance from a dictionary."""
        return cls(
            relation_embedding_dim=config[RELATION_EMBEDDING_DIM],
            scoring_function_norm=config[SCORING_FUNCTION_NORM],
        )


class TransD(BaseModule):
    """An implementation of TransD [ji2015]_.

    This model extends TransR to use fewer parameters.

    .. [ji2015] Ji, G., *et al.* (2015). `Knowledge graph embedding via dynamic mapping matrix
                <http://www.aclweb.org/anthology/P15-1067>`_. ACL.

    .. seealso::

       - Alternative implementation in OpenKE: https://github.com/thunlp/OpenKE/blob/master/models/TransD.py
    """

    model_name = TRANS_D_NAME
    margin_ranking_loss_size_average: bool = True
    entity_embedding_max_norm = 1
    hyper_params = BaseModule.hyper_params + [RELATION_EMBEDDING_DIM, SCORING_FUNCTION_NORM]

    def __init__(self, config: Dict) -> None:
        super().__init__(config)
        config = TransDConfig.from_dict(config)

        # Embeddings
        self.relation_embedding_dim = config.relation_embedding_dim

        # A simple lookup table that stores embeddings of a fixed dictionary and size
        self.relation_embeddings = nn.Embedding(self.num_relations, self.relation_embedding_dim, max_norm=1)
        self.entity_projections = nn.Embedding(self.num_entities, self.embedding_dim)
        self.relation_projections = nn.Embedding(self.num_relations, self.relation_embedding_dim)

        # FIXME @mehdi what about initialization?

        self.scoring_fct_norm = config.scoring_function_norm

    def predict(self, triples):
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)
        scores = self._score_triples(triples)
        return scores.detach().cpu().numpy()

    def _compute_loss(self, positive_scores, negative_scores):
        # y == -1 indicates that second input to criterion should get a larger loss
        # y = torch.Tensor([-1]).cuda()
        # NOTE: y = 1 is important
        # y = torch.tensor([-1], dtype=torch.float, device=self.device)
        y = np.repeat([-1], repeats=positive_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        positive_scores = torch.tensor(positive_scores, dtype=torch.float, device=self.device)
        negative_scores = torch.tensor(negative_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(positive_scores, negative_scores, y)
        return loss

    def forward(self, positives, negatives):
        positive_scores = self._score_triples(positives)
        negative_scores = self._score_triples(negatives)
        loss = self._compute_loss(positive_scores, negative_scores)
        return loss

    def _score_triples(self, triples):
        heads, relations, tails = slice_triples(triples)

        h_embs = self._get_entity_embeddings(heads)
        r_embs = self._get_relation_embeddings(relations)
        t_embs = self._get_entity_embeddings(tails)

        h_proj_vec_embs = self._get_entity_projections(heads)
        r_projs_embs = self._get_relation_projections(relations)
        t_proj_vec_embs = self._get_entity_projections(tails)

        proj_heads = self._project_entities(h_embs, h_proj_vec_embs, r_projs_embs)
        proj_tails = self._project_entities(t_embs, t_proj_vec_embs, r_projs_embs)

        scores = self._compute_scores(h_embs=proj_heads, r_embs=r_embs, t_embs=proj_tails)
        return scores

    def _compute_scores(self, h_embs, r_embs, t_embs):
        # Add the vector element wise
        sum_res = h_embs + r_embs - t_embs
        distances = torch.norm(sum_res, dim=1, p=self.scoring_fct_norm).view(size=(-1,))
        distances = torch.mul(distances, distances)
        return distances

    def _project_entities(self, entity_embs, entity_proj_vecs, relation_projections):
        relation_projections = relation_projections.unsqueeze(-1)
        entity_proj_vecs = entity_proj_vecs.unsqueeze(-1).permute([0, 2, 1])
        transfer_matrices = torch.matmul(relation_projections, entity_proj_vecs)
        projected_entity_embs = torch.einsum('nmk,nk->nm', [transfer_matrices, entity_embs])
        return projected_entity_embs

    def _get_entity_projections(self, entities):
        return self.entity_projections(entities).view(-1, self.embedding_dim)

    def _get_relation_embeddings(self, relations):
        return self.relation_embeddings(relations).view(-1, self.relation_embedding_dim)

    def _get_relation_projections(self, relations):
        return self.relation_projections(relations).view(-1, self.relation_embedding_dim)



# -*- coding: utf-8 -*-

"""Test NDEx access"""

import unittest

from pykeen.utilities.pipeline import load_data


class TestNDExAccess(unittest.TestCase):
    """Test whether NDEx can be accessed."""

    def test_load_graph(self):
        # Load Drugbank
        triples = load_data("ndex:eb1cf70a-1162-11e6-b550-06603eb7f303")

        self.assertIsNotNone(triples)

        # Load BioPlex 2.0
        triples = load_data("ndex:98ba6a19-586e-11e7-8f50-0ac135e8bacf")

        self.assertIsNotNone(triples)



# -*- coding: utf-8 -*-

"""Compute statistics of a KG to be able to interpret the performance of KGE models."""

import operator
from collections import Counter, OrderedDict

import numpy as np


def get_sorted_dict_from_counter(counter):
    """Return sorted dict for Counter object."""
    temp_dict = OrderedDict()

    for key, value in counter.items():
        temp_dict[' '.join(key)] = value

    sorted_x = sorted(temp_dict.items(), key=operator.itemgetter(1))
    sorted_x = OrderedDict(sorted_x)

    return sorted_x


def compute_number_objects_per_subject_relation_tuples(triples):
    """Compute number of objects subject-relation pairs."""
    subjects = triples[:, 0:1]
    relations = triples[:, 1:2]
    subjs_rels = np.concatenate([subjects, relations], axis=-1).tolist()
    temp_tuple = map(tuple, subjs_rels)
    stats = Counter(temp_tuple)

    stats = get_sorted_dict_from_counter(stats)

    return stats


def compute_number_subjects_per_object_relation_tuples(triples):
    """Compute number of subjects per relation-object pairs."""
    objects = triples[:, 2:3]
    relations = triples[:, 1:2]
    objs_rels = np.concatenate([relations, objects], axis=-1).tolist()
    temp_tuple = map(tuple, objs_rels)
    stats = Counter(temp_tuple)

    stats = get_sorted_dict_from_counter(stats)

    return stats




# -*- coding: utf-8 -*-

"""Utilities for getting and initializing KGE models."""

from dataclasses import dataclass
from typing import Dict, Optional, Union

import torch
from torch import nn

from pykeen.constants import (
    EMBEDDING_DIM, GPU, LEARNING_RATE, MARGIN_LOSS, NUM_ENTITIES, NUM_RELATIONS, PREFERRED_DEVICE,
)

__all__ = [
    'BaseModule',
    'BaseConfig',
    'slice_triples',
]


@dataclass
class BaseConfig:
    """Configuration for KEEN models."""

    try_gpu: bool
    margin_loss: float
    number_entities: int
    number_relations: int
    embedding_dimension: int

    def get_device(self):
        """Get the Torch device to use."""
        return torch.device('cuda:0' if torch.cuda.is_available() and self.try_gpu else 'cpu')

    @classmethod
    def from_dict(cls, config: Dict) -> 'BaseConfig':
        """Generate an instance from a dictionary."""
        return cls(
            try_gpu=(config.get(PREFERRED_DEVICE) == GPU),
            margin_loss=config[MARGIN_LOSS],
            number_entities=config[NUM_ENTITIES],
            number_relations=config[NUM_RELATIONS],
            embedding_dimension=config[EMBEDDING_DIM],
        )


class BaseModule(nn.Module):
    """A base class for all of the models."""

    margin_ranking_loss_size_average: bool = ...
    entity_embedding_max_norm: Optional[int] = None
    entity_embedding_norm_type: int = 2
    hyper_params = [EMBEDDING_DIM, MARGIN_LOSS, LEARNING_RATE]

    def __init__(self, config: Union[Dict, BaseConfig]) -> None:
        super().__init__()

        if not isinstance(config, BaseConfig):
            config = BaseConfig.from_dict(config)

        # Device selection
        self.device = config.get_device()

        # Loss
        self.margin_loss = config.margin_loss
        self.criterion = nn.MarginRankingLoss(
            margin=self.margin_loss,
            size_average=self.margin_ranking_loss_size_average,
        )

        # Entity dimensions
        #: The number of entities in the knowledge graph
        self.num_entities = config.number_entities
        #: The number of unique relation types in the knowledge graph
        self.num_relations = config.number_relations
        #: The dimension of the embeddings to generate
        self.embedding_dim = config.embedding_dimension

        self.entity_embeddings = nn.Embedding(
            self.num_entities,
            self.embedding_dim,
            norm_type=self.entity_embedding_norm_type,
            max_norm=self.entity_embedding_max_norm,
        )

    def __init_subclass__(cls, **kwargs):  # noqa: D105
        if not getattr(cls, 'model_name', None):
            raise TypeError('missing model_name class attribute')

    def _get_entity_embeddings(self, entities):
        return self.entity_embeddings(entities).view(-1, self.embedding_dim)


def slice_triples(triples):
    """Get the heads, relations, and tails from a matrix of triples."""
    h = triples[:, 0:1]
    r = triples[:, 1:2]
    t = triples[:, 2:3]
    return h, r, t



# -*- coding: utf-8 -*-

"""Implementation of ConvE."""

from typing import Dict

import torch
import torch.autograd
from torch import nn
from torch.nn import Parameter, functional as F
from torch.nn.init import xavier_normal

from pykeen.constants import (
    CONV_E_FEATURE_MAP_DROPOUT, CONV_E_HEIGHT, CONV_E_INPUT_CHANNELS, CONV_E_INPUT_DROPOUT, CONV_E_KERNEL_HEIGHT,
    CONV_E_KERNEL_WIDTH, CONV_E_NAME, CONV_E_OUTPUT_CHANNELS, CONV_E_OUTPUT_DROPOUT, CONV_E_WIDTH, EMBEDDING_DIM,
    NUM_ENTITIES, NUM_RELATIONS,
    MARGIN_LOSS, LEARNING_RATE)

__all__ = ['ConvE']


class ConvE(nn.Module):
    """An implementation of ConvE [dettmers2017]_.

    .. [dettmers2017] Dettmers, T., *et al.* (2017) `Convolutional 2d knowledge graph embeddings
                      <https://arxiv.org/pdf/1707.01476.pdf>`_. arXiv preprint arXiv:1707.01476.

    .. seealso:: https://github.com/TimDettmers/ConvE/blob/master/model.py
    """

    model_name = CONV_E_NAME
    hyper_params = [EMBEDDING_DIM, CONV_E_INPUT_CHANNELS, CONV_E_OUTPUT_CHANNELS, CONV_E_HEIGHT, CONV_E_WIDTH,
                    CONV_E_KERNEL_HEIGHT, CONV_E_KERNEL_WIDTH, CONV_E_INPUT_DROPOUT, CONV_E_FEATURE_MAP_DROPOUT,
                    CONV_E_OUTPUT_DROPOUT, MARGIN_LOSS, LEARNING_RATE]

    def __init__(self, config: Dict) -> None:
        super().__init__()

        # Device selection
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        # Entity dimensions
        self.num_entities = config[NUM_ENTITIES]
        self.num_relations = config[NUM_RELATIONS]

        # Embeddings
        self.embedding_dim = config[EMBEDDING_DIM]

        self.entity_embeddings = nn.Embedding(self.num_entities, self.embedding_dim)
        self.relation_embeddings = nn.Embedding(self.num_relations, self.embedding_dim)

        num_in_channels = config[CONV_E_INPUT_CHANNELS]
        num_out_channels = config[CONV_E_OUTPUT_CHANNELS]
        self.img_height = config[CONV_E_HEIGHT]
        self.img_width = config[CONV_E_WIDTH]
        kernel_height = config[CONV_E_KERNEL_HEIGHT]
        kernel_width = config[CONV_E_KERNEL_WIDTH]
        input_dropout = config[CONV_E_INPUT_DROPOUT]
        hidden_dropout = config[CONV_E_OUTPUT_DROPOUT]
        feature_map_dropout = config[CONV_E_FEATURE_MAP_DROPOUT]

        assert self.img_height * self.img_width == self.embedding_dim

        self.inp_drop = torch.nn.Dropout(input_dropout)
        self.hidden_drop = torch.nn.Dropout(hidden_dropout)
        self.feature_map_drop = torch.nn.Dropout2d(feature_map_dropout)
        self.loss = torch.nn.BCELoss()

        self.conv1 = torch.nn.Conv2d(
            in_channels=num_in_channels,
            out_channels=num_out_channels,
            kernel_size=(kernel_height, kernel_width),
            stride=1,
            padding=0,
            bias=True,
        )

        # num_features – C from an expected input of size (N,C,L)
        self.bn0 = torch.nn.BatchNorm2d(num_in_channels)
        # num_features – C from an expected input of size (N,C,H,W)
        self.bn1 = torch.nn.BatchNorm2d(num_out_channels)
        self.bn2 = torch.nn.BatchNorm1d(self.embedding_dim)
        self.register_parameter('b', Parameter(torch.zeros(self.num_entities)))
        num_in_features = num_out_channels * \
                          (2 * self.img_height - kernel_height + 1) * \
                          (self.img_width - kernel_width + 1)
        self.fc = torch.nn.Linear(num_in_features, self.embedding_dim)

    def init(self):  # FIXME is this ever called?
        xavier_normal(self.entity_embeddings.weight.data)
        xavier_normal(self.relation_embeddings.weight.data)

    def predict(self, triple_batch):
        batch_size = triple_batch.shape[0]
        triple_batch = torch.tensor(triple_batch, dtype=torch.long, device=self.device)
        subject_batch = triple_batch[:, 0:1]
        relation_batch = triple_batch[:, 1:2]
        object_batch = triple_batch[:, 2:3].view(-1)

        subject_batch_embedded = self.entity_embeddings(subject_batch).view(-1, 1, self.img_height, self.img_width)
        relation_batch_embedded = self.entity_embeddings(relation_batch).view(-1, 1, self.img_height, self.img_width)
        candidate_object_emebddings = self.entity_embeddings(object_batch)

        # batch_size, num_input_channels, 2*height, width
        stacked_inputs = torch.cat([subject_batch_embedded, relation_batch_embedded], 2)

        # batch_size, num_input_channels, 2*height, width
        stacked_inputs = self.bn0(stacked_inputs)

        # batch_size, num_input_channels, 2*height, width
        x = self.inp_drop(stacked_inputs)
        # (N,C_out,H_out,W_out)
        x = self.conv1(x)

        x = self.bn1(x)
        x = F.relu(x)
        x = self.feature_map_drop(x)
        # batch_size, num_output_channels * (2 * height - kernel_height + 1) * (width - kernel_width + 1)
        x = x.view(batch_size, -1)
        x = self.fc(x)
        x = self.hidden_drop(x)

        if batch_size > 1:
            x = self.bn2(x)
        x = F.relu(x)
        x = torch.mm(x, candidate_object_emebddings.transpose(1, 0))
        scores = F.sigmoid(x)

        max_scores, predicted_entities = torch.max(scores, dim=1)

        return predicted_entities

    def forward(self, batch, labels):
        batch_size = batch.shape[0]

        heads = batch[:, 0:1]
        relations = batch[:, 1:2]
        tails = batch[:, 2:3]

        # batch_size, num_input_channels, width, height
        heads_embs = self.entity_embeddings(heads).view(-1, 1, self.img_height, self.img_width)
        relation_embs = self.relation_embeddings(relations).view(-1, 1, self.img_height, self.img_width)
        tails_embs = self.entity_embeddings(tails).view(-1, self.embedding_dim)

        # batch_size, num_input_channels, 2*height, width
        stacked_inputs = torch.cat([heads_embs, relation_embs], 2)

        # batch_size, num_input_channels, 2*height, width
        stacked_inputs = self.bn0(stacked_inputs)

        # batch_size, num_input_channels, 2*height, width
        x = self.inp_drop(stacked_inputs)
        # (N,C_out,H_out,W_out)
        x = self.conv1(x)

        x = self.bn1(x)
        x = F.relu(x)
        x = self.feature_map_drop(x)
        # batch_size, num_output_channels * (2 * height - kernel_height + 1) * (width - kernel_width + 1)
        x = x.view(batch_size, -1)
        x = self.fc(x)
        x = self.hidden_drop(x)

        if batch_size > 1:
            x = self.bn2(x)
        x = F.relu(x)

        scores = torch.sum(torch.mm(x, tails_embs.transpose(1, 0)), dim=1)

        predictions = F.sigmoid(scores)
        loss = self.loss(predictions, labels)
        return loss



# -*- coding: utf-8 -*-

"""PyKEEN's command line interface."""

from pykeen.cli.utils import (
    configure_conv_e_hpo_pipeline, configure_conv_e_training_pipeline,
    configure_distmult_hpo_pipeline, configure_distmult_training_pipeline, configure_ermlp_hpo_pipeline,
    configure_ermlp_training_pipeline, configure_rescal_hpo_pipeline, configure_rescal_training_pipeline,
    configure_se_hpo_pipeline, configure_se_training_pipeline, configure_trans_d_hpo_pipeline,
    configure_trans_d_training_pipeline, configure_trans_e_hpo_pipeline, configure_trans_e_training_pipeline,
    configure_trans_h_hpo_pipeline, configure_trans_h_training_pipeline, configure_trans_r_hpo_pipeline,
    configure_trans_r_training_pipeline, configure_um_hpo_pipeline, configure_um_training_pipeline,
)
from pykeen.kge_models import (
    ConvE, DistMult, ERMLP, RESCAL, StructuredEmbedding, TransD, TransE, TransH, TransR, UnstructuredModel,
)

__all__ = [
    'MODEL_TRAINING_CONFIG_FUNCS',
    'MODEL_HPO_CONFIG_FUNCS',
]

MODEL_TRAINING_CONFIG_FUNCS = {
    TransE.model_name: configure_trans_e_training_pipeline,
    TransH.model_name: configure_trans_h_training_pipeline,
    TransR.model_name: configure_trans_r_training_pipeline,
    TransD.model_name: configure_trans_d_training_pipeline,
    StructuredEmbedding.model_name: configure_se_training_pipeline,
    UnstructuredModel.model_name: configure_um_training_pipeline,
    DistMult.model_name: configure_distmult_training_pipeline,
    ERMLP.model_name: configure_ermlp_training_pipeline,
    RESCAL.model_name: configure_rescal_training_pipeline,
    ConvE.model_name: configure_conv_e_training_pipeline,
}

MODEL_HPO_CONFIG_FUNCS = {
    TransE.model_name: configure_trans_e_hpo_pipeline,
    TransH.model_name: configure_trans_h_hpo_pipeline,
    TransR.model_name: configure_trans_r_hpo_pipeline,
    TransD.model_name: configure_trans_d_hpo_pipeline,
    StructuredEmbedding.model_name: configure_se_hpo_pipeline,
    UnstructuredModel.model_name: configure_um_hpo_pipeline,
    DistMult.model_name: configure_distmult_hpo_pipeline,
    ERMLP.model_name: configure_ermlp_hpo_pipeline,
    RESCAL.model_name: configure_rescal_hpo_pipeline,
    ConvE.model_name: configure_conv_e_hpo_pipeline,
}



# -*- coding: utf-8 -*-

"""Implementation the command line interface needed for Structured Model (SM)."""

from collections import OrderedDict

import pykeen.constants as pkc
from pykeen.cli.utils.cli_print_msg_helper import (
    print_batch_size_message, print_embedding_dimension_info_message, print_entity_normalization_message,
    print_hpo_batch_sizes_message, print_hpo_embedding_dimensions_message, print_hpo_entity_normalization_norms_message,
    print_hpo_epochs_message, print_hpo_learning_rates_message, print_hpo_margin_losses_message,
    print_hpo_scoring_fcts_message, print_learning_rate_message, print_number_epochs_message, print_scoring_fct_message,
    print_section_divider, print_training_embedding_dimension_message, print_training_margin_loss_message,
)
from pykeen.cli.utils.cli_query_helper import (
    select_float_value, select_float_values, select_integer_value, select_positive_integer_values,
)
from pykeen.utilities.summarize import get_config_dict

__all__ = [
    'configure_se_hpo_pipeline',
    'configure_se_training_pipeline',
]


def configure_se_training_pipeline(model_name: str):
    config = OrderedDict()
    config[pkc.KG_EMBEDDING_MODEL_NAME] = model_name

    # Step 1: Query embedding dimension
    print_training_embedding_dimension_message()
    print_embedding_dimension_info_message()
    embedding_dimension = select_integer_value(
        print_msg=pkc.EMBEDDING_DIMENSION_PRINT_MSG,
        prompt_msg=pkc.EMBEDDING_DIMENSION_PROMPT_MSG,
        error_msg=pkc.EMBEDDING_DIMENSION_ERROR_MSG,
    )
    config[pkc.EMBEDDING_DIM] = embedding_dimension
    print_section_divider()

    # Step 2: Query margin loss
    print_training_margin_loss_message()
    magin_loss = select_float_value(
        print_msg=pkc.MARGIN_LOSS_PRINT_MSG,
        prompt_msg=pkc.MARGIN_LOSS_PROMPT_MSG,
        error_msg=pkc.MARGIN_LOSS_ERROR_MSG,
    )
    config[pkc.MARGIN_LOSS] = magin_loss
    print_section_divider()

    # Step 3: Query L_p norm as scoring function
    print_scoring_fct_message()
    scoring_fct_norm = select_integer_value(
        print_msg=pkc.NORM_SCORING_FUNCTION_PRINT_MSG,
        prompt_msg=pkc.NORM_SCORING_FUNCTION_PROMPT_MSG,
        error_msg=pkc.NORM_SCORING_FUNCTION_ERROR_MSG,
    )
    config[pkc.SCORING_FUNCTION_NORM] = scoring_fct_norm
    print_section_divider()

    # Step 4: Query L_p norm for normalizing the entities
    print_entity_normalization_message()
    entity_normalization_norm = select_integer_value(
        print_msg=pkc.ENTITIES_NORMALIZATION_PRINT_MSG,
        prompt_msg=pkc.ENTITIES_NORMALIZATION_PROMPT_MSG,
        error_msg=pkc.ENTITIES_NORMALIZATION_ERROR_MSG,
    )
    config[pkc.NORM_FOR_NORMALIZATION_OF_ENTITIES] = entity_normalization_norm
    print_section_divider()

    # Step 5: Query learning rate
    print_learning_rate_message()
    learning_rate = select_float_value(
        print_msg=pkc.LEARNING_RATE_PRINT_MSG,
        prompt_msg=pkc.LEARNING_RATE_PROMPT_MSG,
        error_msg=pkc.LEARNING_RATE_ERROR_MSG,
    )
    config[pkc.LEARNING_RATE] = learning_rate
    print_section_divider()

    # Step 6: Query batch size
    print_batch_size_message()
    batch_size = select_integer_value(
        print_msg=pkc.BATCH_SIZE_PRINT_MSG,
        prompt_msg=pkc.BATCH_SIZE_PROMPT_MSG,
        error_msg=pkc.BATCH_SIZE_ERROR_MSG,
    )
    config[pkc.BATCH_SIZE] = batch_size
    print_section_divider()

    # Step 7: Query number of epochs
    print_number_epochs_message()
    number_epochs = select_integer_value(
        print_msg=pkc.EPOCH_PRINT_MSG,
        prompt_msg=pkc.EPOCH_PROMPT_MSG,
        error_msg=pkc.EPOCH_ERROR_MSG,
    )
    config[pkc.NUM_EPOCHS] = number_epochs
    print_section_divider()

    return config


def configure_se_hpo_pipeline(model_name):
    config = get_config_dict(model_name)

    # Step 1: Query embedding dimensions
    print_hpo_embedding_dimensions_message()
    embedding_dimensions = select_positive_integer_values(
        pkc.EMBEDDING_DIMENSIONS_PRINT_MSG,
        pkc.EMBEDDING_DIMENSIONS_PROMPT_MSG,
        pkc.EMBEDDING_DIMENSIONS_ERROR_MSG,
    )
    config[pkc.EMBEDDING_DIM] = embedding_dimensions
    print_section_divider()

    # Step 2: Query margin loss
    print_hpo_margin_losses_message()
    magin_loss = select_float_values(
        print_msg=pkc.MARGIN_LOSSES_PRINT_MSG,
        prompt_msg=pkc.MARGIN_LOSSES_PROMPT_MSG,
        error_msg=pkc.MARGIN_LOSSES_ERROR_MSG,
    )

    config[pkc.MARGIN_LOSS] = magin_loss
    print_section_divider()

    # Step 3: Query L_p norms to use as scoring function
    print_hpo_scoring_fcts_message()
    scoring_fct_norm = select_positive_integer_values(
        print_msg=pkc.NORMS_SCORING_FUNCTION_PRINT_MSG,
        prompt_msg=pkc.NORMS_SCORING_FUNCTION_PROMPT_MSG,
        error_msg=pkc.NORMS_SCORING_FUNCTION_ERROR_MSG,
    )
    config[pkc.SCORING_FUNCTION_NORM] = scoring_fct_norm
    print_section_divider()

    # Step 4: Query L_p norms for normalizing the entities
    print_hpo_entity_normalization_norms_message()
    entity_normalization_norm = select_positive_integer_values(
        print_msg=pkc.NORMS_FOR_NORMALIZATION_OF_ENTITIES_PRINT_MSG,
        prompt_msg=pkc.NORMS_FOR_NORMALIZATION_OF_ENTITIES_PROMPT_MSG,
        error_msg=pkc.NORMS_FOR_NORMALIZATION_OF_ENTITIES_ERROR_MSG,
    )
    config[pkc.NORM_FOR_NORMALIZATION_OF_ENTITIES] = entity_normalization_norm
    print_section_divider()

    # Step 5: Query learning rate
    print_hpo_learning_rates_message()
    learning_rate = select_float_values(
        print_msg=pkc.LEARNING_RATES_PRINT_MSG,
        prompt_msg=pkc.LEARNING_RATES_PROMPT_MSG,
        error_msg=pkc.LEARNING_RATES_ERROR_MSG,
    )
    config[pkc.LEARNING_RATE] = learning_rate
    print_section_divider()

    # Step 6: Query batch size
    print_hpo_batch_sizes_message()
    batch_size = select_positive_integer_values(
        print_msg=pkc.BATCH_SIZES_PRINT_MSG,
        prompt_msg=pkc.BATCH_SIZES_PROMPT_MSG,
        error_msg=pkc.BATCH_SIZES_ERROR_MSG,
    )
    config[pkc.BATCH_SIZE] = batch_size
    print_section_divider()

    # Step 7: Query number of epochs
    print_hpo_epochs_message()
    number_epochs = select_positive_integer_values(
        print_msg=pkc.EPOCHS_PRINT_MSG,
        prompt_msg=pkc.EPOCHS_PROMPT_MSG,
        error_msg=pkc.EPOCHS_ERROR_MSG,
    )
    config[pkc.NUM_EPOCHS] = number_epochs
    print_section_divider()

    return config



# -*- coding: utf-8 -*-

"""A hyper-parameter optimizer that uses random search."""

import random
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import torch
from torch.nn import Module
from tqdm import trange

import pykeen.constants as pkc
from pykeen.hyper_parameter_optimizer.abstract_hyper_params_optimizer import AbstractHPOptimizer
from pykeen.kge_models import get_kge_model
from pykeen.utilities.evaluation_utils.metrics_computations import compute_metric_results
from pykeen.utilities.train_utils import train_kge_model

__all__ = [
    'RandomSearchHPO',
]

OptimizeResult = Tuple[Module, List[float], Any, Any, Any, Any]


class RandomSearchHPO(AbstractHPOptimizer):
    """A hyper-parameter optimizer that uses random search."""

    def _sample_conv_e_params(self, hyperparams_dict) -> Dict[str, Any]:
        kg_model_config = OrderedDict()
        # Sample params which are dependent on each other
        embedding_dimensions = hyperparams_dict[pkc.EMBEDDING_DIM]
        sampled_index = random.choice(range(len(embedding_dimensions)))
        kg_model_config[pkc.EMBEDDING_DIM] = embedding_dimensions[sampled_index]
        kg_model_config[pkc.CONV_E_HEIGHT] = hyperparams_dict[pkc.CONV_E_HEIGHT][sampled_index]
        kg_model_config[pkc.CONV_E_WIDTH] = hyperparams_dict[pkc.CONV_E_WIDTH][sampled_index]
        kg_model_config[pkc.CONV_E_KERNEL_HEIGHT] = hyperparams_dict[pkc.CONV_E_KERNEL_HEIGHT][sampled_index]
        kg_model_config[pkc.CONV_E_KERNEL_WIDTH] = hyperparams_dict[pkc.CONV_E_KERNEL_WIDTH][sampled_index]

        del hyperparams_dict[pkc.EMBEDDING_DIM]
        del hyperparams_dict[pkc.CONV_E_HEIGHT]
        del hyperparams_dict[pkc.CONV_E_WIDTH]
        del hyperparams_dict[pkc.CONV_E_KERNEL_HEIGHT]
        del hyperparams_dict[pkc.CONV_E_KERNEL_WIDTH]

        kg_model_config.update(self._sample_parameter_value(hyperparams_dict))

        return kg_model_config

    def optimize_hyperparams(self,
                             mapped_train_triples,
                             mapped_test_triples,
                             entity_to_id,
                             rel_to_id,
                             config,
                             device,
                             seed: Optional[int] = None,
                             k_evaluation: int = 10) -> OptimizeResult:
        """"""
        if seed is not None:
            torch.manual_seed(config[pkc.SEED])

        trained_kge_models: List[Module] = []
        epoch_losses: List[List[float]] = []
        hits_at_k_evaluations: List[float] = []
        entity_to_ids: List[Dict[int, str]] = []
        rel_to_ids: List[Dict[int, str]] = []
        models_params: List[Dict] = []
        eval_summaries: List = []

        config = config.copy()
        max_iters = config[pkc.NUM_OF_HPO_ITERS]

        sample_fct = (
            self._sample_conv_e_params
            if config[pkc.KG_EMBEDDING_MODEL_NAME] == pkc.CONV_E_NAME else
            self._sample_parameter_value
        )

        for _ in trange(max_iters, desc='HPO Iteration'):
            # Sample hyper-params
            kge_model_config: Dict[str, Any] = sample_fct(config)
            kge_model_config[pkc.NUM_ENTITIES]: int = len(entity_to_id)
            kge_model_config[pkc.NUM_RELATIONS]: int = len(rel_to_id)
            kge_model_config[pkc.SEED]: int = seed

            # Configure defined model
            kge_model: Module = get_kge_model(config=kge_model_config)

            models_params.append(kge_model_config)
            entity_to_ids.append(entity_to_id)
            rel_to_ids.append(rel_to_id)

            all_entities = np.array(list(entity_to_id.values()))

            trained_kge_model, epoch_loss = train_kge_model(
                kge_model=kge_model,
                all_entities=all_entities,
                learning_rate=kge_model_config[pkc.LEARNING_RATE],
                num_epochs=kge_model_config[pkc.NUM_EPOCHS],
                batch_size=kge_model_config[pkc.BATCH_SIZE],
                pos_triples=mapped_train_triples,
                device=device,
                seed=seed,
                tqdm_kwargs=dict(leave=False),
            )

            # Evaluate trained model
            metric_results = compute_metric_results(
                all_entities=all_entities,
                kg_embedding_model=trained_kge_model,
                mapped_train_triples=mapped_train_triples,
                mapped_test_triples=mapped_test_triples,
                device=device,
            )

            # TODO: Define HPO metric
            eval_summaries.append(metric_results)

            trained_kge_models.append(trained_kge_model)
            epoch_losses.append(epoch_loss)

            hits_at_k_evaluation = metric_results.hits_at_k[k_evaluation]
            hits_at_k_evaluations.append(hits_at_k_evaluation)

        index_of_max = int(np.argmax(a=hits_at_k_evaluations))

        return (
            trained_kge_models[index_of_max],
            epoch_losses[index_of_max],
            entity_to_ids[index_of_max],
            rel_to_ids[index_of_max],
            eval_summaries[index_of_max],
            models_params[index_of_max],
        )

    @classmethod
    def run(cls,
            mapped_train_triples: np.ndarray,
            mapped_test_triples: np.ndarray,
            entity_to_id: Dict[int, str],
            rel_to_id: Dict[int, str],
            config: Dict,
            device,
            seed) -> OptimizeResult:
        return cls().optimize_hyperparams(
            mapped_train_triples=mapped_train_triples,
            mapped_test_triples=mapped_test_triples,
            entity_to_id=entity_to_id,
            rel_to_id=rel_to_id,
            config=config,
            device=device,
            seed=seed,
        )




# -*- coding: utf-8 -*-

"""Reading of Freebase and related data."""

__all__ = [
    'transform_freebase_to_internal_format',
    'transform_wn_18_to_internal_format',
]


def transform_freebase_to_internal_format(data_path_in: str, data_path_out: str) -> None:
    _transform_spo_to_internal_format(data_path_in, data_path_out)


def transform_wn_18_to_internal_format(data_path_in: str, data_path_out: str) -> None:
    _transform_spo_to_internal_format(data_path_in, data_path_out)


def _transform_spo_to_internal_format(data_path_in: str, data_path_out: str) -> None:
    with open(data_path_in, 'r', encoding='utf-8') as file_in, open(data_path_out, 'w', encoding='utf-8') as file_out:
        print('@Comment@ Subject Predicate Object', file=file_out)
        for line in file_in:
            s, p, o = line.strip().split('\t')
            print(s, p, o, sep='\t', file=file_out)



# -*- coding: utf-8 -*-

"""Utilities for the command line interface."""

from .conv_e_cli import configure_conv_e_hpo_pipeline, configure_conv_e_training_pipeline
from .distmult_cli import configure_distmult_hpo_pipeline, configure_distmult_training_pipeline
from .ermlp_cli import configure_ermlp_hpo_pipeline, configure_ermlp_training_pipeline
from .rescal_cli import configure_rescal_hpo_pipeline, configure_rescal_training_pipeline
from .structured_embedding_cli import configure_se_hpo_pipeline, configure_se_training_pipeline
from .trans_d_cli import configure_trans_d_hpo_pipeline, configure_trans_d_training_pipeline
from .trans_e_cli import configure_trans_e_hpo_pipeline, configure_trans_e_training_pipeline
from .trans_h_cli import configure_trans_h_hpo_pipeline, configure_trans_h_training_pipeline
from .trans_r_cli import configure_trans_r_hpo_pipeline, configure_trans_r_training_pipeline
from .unstructured_model_cli import configure_um_hpo_pipeline, configure_um_training_pipeline

__all__ = [
    'configure_conv_e_hpo_pipeline', 'configure_conv_e_training_pipeline',
    'configure_distmult_hpo_pipeline', 'configure_distmult_training_pipeline',
    'configure_ermlp_hpo_pipeline', 'configure_ermlp_training_pipeline',
    'configure_rescal_hpo_pipeline', 'configure_rescal_training_pipeline',
    'configure_se_hpo_pipeline', 'configure_se_training_pipeline',
    'configure_trans_d_hpo_pipeline', 'configure_trans_d_training_pipeline',
    'configure_trans_e_hpo_pipeline', 'configure_trans_e_training_pipeline',
    'configure_trans_h_hpo_pipeline', 'configure_trans_h_training_pipeline',
    'configure_trans_r_hpo_pipeline', 'configure_trans_r_training_pipeline',
    'configure_um_hpo_pipeline', 'configure_um_training_pipeline',
]



# -*- coding: utf-8 -*-

"""PyKEEN's command line interface."""

import json
import os
from typing import TextIO

import click
from click_default_group import DefaultGroup

from pykeen.cli.prompt import prompt_config
from pykeen.predict import start_predictions_pipeline
from pykeen.run import run
from pykeen.utilities.summarize import summarize_results


@click.group(cls=DefaultGroup, default='train', default_if_no_args=True)
@click.version_option()
def main():
    """PyKEEN: A software for training and evaluating knowledge graph embeddings."""


@main.command()
@click.option('-c', '--config', type=click.File(), help='A PyKEEN JSON configuration file')
def train(config):
    """Train a KGE model."""
    if config is not None:
        config = json.load(config)
    else:
        config = prompt_config()

    run(config)


@main.command()
@click.option('-m', '--model-directory', type=click.Path(file_okay=False, dir_okay=True, exists=True))
@click.option('-d', '--data-directory', type=click.Path(file_okay=False, dir_okay=True))
@click.option('-t', '--training-set-path', type=click.Path(file_okay=True, dir_okay=False, exists=True))
def predict(model_directory: str, data_directory: str, training_set_path: str):
    """Predict new links based on trained model."""
    start_predictions_pipeline(
        model_directory,
        data_directory,
        remove_training_triples=(training_set_path is not None),
        training_set_path=training_set_path,
    )


@main.command()
@click.option('-d', '--directory', type=click.Path(file_okay=False, dir_okay=True), default=os.getcwd())
@click.option('-o', '--output', type=click.File('w'), required=True)
def summarize(directory: str, output: TextIO):
    """Summarize contents of training and evaluation."""
    summarize_results(directory, output)


@main.command()
def prefixes():
    """List registered prefixes."""
    from pykeen.constants import IMPORTERS
    for name, handler in sorted(IMPORTERS.items()):
        handler_doc = handler.__doc__.split('\n')[0]
        click.echo(f'{name}: {handler_doc}')


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Script for starting the pipeline and saving the results."""

import json
import os
import pickle
import time
from dataclasses import dataclass
from typing import Dict, Mapping, Optional

import numpy as np
import torch

from pykeen.constants import (
    ENTITY_TO_EMBEDDING, ENTITY_TO_ID, EVAL_SUMMARY, FINAL_CONFIGURATION, LOSSES, OUTPUT_DIREC, RELATION_TO_EMBEDDING,
    RELATION_TO_ID, TRAINED_MODEL,
)
from pykeen.utilities.pipeline import Pipeline

__all__ = [
    'Results',
    'run',
]


@dataclass
class Results:
    """Results from PyKEEN."""

    #: The configuration used to train the KGE model
    config: Mapping

    #: The pipeline used to train the KGE model
    pipeline: Pipeline

    #: The results of training the KGE model
    results: Mapping

    @property
    def trained_model(self) -> torch.nn.Module:  # noqa: D401
        """The pre-trained KGE model."""
        return self.results['trained_model']

    @property
    def losses(self):  # noqa: D401
        """The losses calculated during training."""
        return self.results['losses']

    @property
    def evaluation_summary(self):  # noqa: D401
        """The evaluation summary."""
        return self.results['eval_summary']

    def plot_losses(self) -> None:
        """Plot the losses using Matplotlib."""
        import matplotlib.pyplot as plt
        epochs = np.arange(len(self.losses))
        plt.title('Loss Per Epoch')
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.plot(epochs, self.losses)


def run(config: Dict,
        output_directory: Optional[str] = None,
        ) -> Results:
    """Train a KGE model.

    :param config: The configuration specifying the KGE model and its hyper-parameters
    :param output_directory: The directory to store the results
    """
    if output_directory is None:
        output_directory = os.path.join(config[OUTPUT_DIREC], time.strftime("%Y-%m-%d-%H-%M-%S"))
    os.makedirs(output_directory, exist_ok=True)

    pipeline = Pipeline(config=config)
    pipeline_results = pipeline.run()

    with open(os.path.join(output_directory, 'configuration.json'), 'w') as file:
        # In HPO model initial configuration is different from final configurations, that's why we differentiate
        json.dump(pipeline_results[FINAL_CONFIGURATION], file, indent=2)

    with open(os.path.join(output_directory, 'entities_to_embeddings.pkl'), 'wb') as file:
        pickle.dump(pipeline_results[ENTITY_TO_EMBEDDING], file, protocol=pickle.HIGHEST_PROTOCOL)

    with open(os.path.join(output_directory, 'entities_to_embeddings.json'), 'w') as file:
        json.dump(
            {
                key: list(map(float, array))
                for key, array in pipeline_results[ENTITY_TO_EMBEDDING].items()
            },
            file,
            indent=2,
            sort_keys=True,
        )

    if pipeline_results[RELATION_TO_EMBEDDING] is not None:
        with open(os.path.join(output_directory, 'relations_to_embeddings.pkl'), 'wb') as file:
            pickle.dump(pipeline_results[RELATION_TO_EMBEDDING], file, protocol=pickle.HIGHEST_PROTOCOL)

        with open(os.path.join(output_directory, 'relations_to_embeddings.json'), 'w') as file:
            json.dump(
                {
                    key: list(map(float, array))
                    for key, array in pipeline_results[RELATION_TO_EMBEDDING].items()
                },
                file,
                indent=2,
                sort_keys=True,
            )

    with open(os.path.join(output_directory, 'entity_to_id.json'), 'w') as file:
        json.dump(pipeline_results[ENTITY_TO_ID], file, indent=2, sort_keys=True)

    with open(os.path.join(output_directory, 'relation_to_id.json'), 'w') as file:
        json.dump(pipeline_results[RELATION_TO_ID], file, indent=2, sort_keys=True)

    with open(os.path.join(output_directory, 'losses.json'), 'w') as file:
        json.dump(pipeline_results[LOSSES], file, indent=2, sort_keys=True)

    eval_summary = pipeline_results.get(EVAL_SUMMARY)
    if eval_summary is not None:
        with open(os.path.join(output_directory, 'evaluation_summary.json'), 'w') as file:
            json.dump(eval_summary, file, indent=2)

    # Save trained model
    torch.save(
        pipeline_results[TRAINED_MODEL].state_dict(),
        os.path.join(output_directory, 'trained_model.pkl'),
    )

    return Results(
        config=config,
        pipeline=pipeline,
        results=pipeline_results,
    )



# -*- coding: utf-8 -*-

"""Implementation of UM."""

import logging
from typing import Dict

import numpy as np
import torch
import torch.autograd
from torch import nn

from pykeen.constants import NORM_FOR_NORMALIZATION_OF_ENTITIES, SCORING_FUNCTION_NORM, UM_NAME
from pykeen.kge_models.base import BaseModule
from .trans_e import TransEConfig

__all__ = ['UnstructuredModel']

log = logging.getLogger(__name__)


class UnstructuredModel(BaseModule):
    """An implementation of Unstructured Model (UM) [bordes2014]_.

    .. [bordes2014] Bordes, A., *et al.* (2014). `A semantic matching energy function for learning with
                    multi-relational data <https://link.springer.com/content/pdf/10.1007%2Fs10994-013-5363-6.pdf>`_.
                    Machine
    """

    model_name = UM_NAME
    margin_ranking_loss_size_average: bool = True
    hyper_params = BaseModule.hyper_params + [SCORING_FUNCTION_NORM, NORM_FOR_NORMALIZATION_OF_ENTITIES]

    def __init__(self, config: Dict) -> None:
        super().__init__(config)
        config = TransEConfig.from_dict(config)

        self.l_p_norm_entities = config.lp_norm
        self.scoring_fct_norm = config.scoring_function_norm

        self._initialize()

    def _initialize(self):
        entity_embeddings_init_bound = 6 / np.sqrt(self.embedding_dim)
        nn.init.uniform_(
            self.entity_embeddings.weight.data,
            a=-entity_embeddings_init_bound,
            b=entity_embeddings_init_bound,
        )

    def predict(self, triples):
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)
        scores = self._score_triples(triples)
        return scores.detach().cpu().numpy()

    def forward(self, batch_positives, batch_negatives):
        # Normalize embeddings of entities
        pos_scores = self._score_triples(batch_positives)
        neg_scores = self._score_triples(batch_negatives)
        loss = self._compute_loss(pos_scores=pos_scores, neg_scores=neg_scores)
        return loss

    def _compute_loss(self, pos_scores, neg_scores):
        y = np.repeat([-1], repeats=pos_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        pos_scores = torch.tensor(pos_scores, dtype=torch.float, device=self.device)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float, device=self.device)
        # neg_scores_temp = 1 * torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(pos_scores, neg_scores, y)
        return loss

    def _score_triples(self, triples):
        head_embeddings, tail_embeddings = self._get_triple_embeddings(triples)
        scores = self._compute_scores(head_embeddings=head_embeddings, tail_embeddings=tail_embeddings)
        return scores

    def _compute_scores(self, head_embeddings, tail_embeddings):
        # Add the vector element wise
        sum_res = head_embeddings - tail_embeddings
        distances = torch.norm(sum_res, dim=1, p=self.scoring_fct_norm).view(size=(-1,))
        distances = distances ** 2
        return distances

    def _get_triple_embeddings(self, triples):
        heads, tails = self.slice_triples(triples)
        return (
            self._get_entity_embeddings(heads),
            self._get_entity_embeddings(tails),
        )

    def _get_entity_embeddings(self, entities):
        return self.entity_embeddings(entities).view(-1, self.embedding_dim)

    @staticmethod
    def slice_triples(triples):
        return (
            triples[:, 0:1],
            triples[:, 2:3],
        )



# -*- coding: utf-8 -*-

"""Implementation of TransR."""

from typing import Dict

import numpy as np
import torch
import torch.autograd
from torch import nn

from pykeen.constants import RELATION_EMBEDDING_DIM, SCORING_FUNCTION_NORM, TRANS_R_NAME
from .base import BaseModule
from .trans_d import TransDConfig

__all__ = ['TransR']


class TransR(BaseModule):
    """An implementation of TransR [lin2015]_.

    This model extends TransE and TransH by considering different vector spaces for entities and relations.

    .. [lin2015] Lin, Y., *et al.* (2015). `Learning entity and relation embeddings for knowledge graph completion
                 <http://www.aaai.org/ocs/index.php/AAAI/AAAI15/paper/download/9571/9523/>`_. AAAI. Vol. 15.

    Constraints:
     * ||h||_2 <= 1: Done
     * ||r||_2 <= 1: Done
     * ||t||_2 <= 1: Done
     * ||h*M_r||_2 <= 1: Done
     * ||t*M_r||_2 <= 1: Done

    .. seealso::

       - Implementation in OpenKE: https://github.com/thunlp/OpenKE/blob/master/models/TransR.py
       - PyTorch implementation in OpenKE: https://github.com/thunlp/OpenKE/blob/OpenKE-PyTorch/models/TransR.py
    """

    model_name = TRANS_R_NAME
    margin_ranking_loss_size_average: bool = True
    # TODO: max_norm < 1.
    # max_norm = 1 according to the paper
    entity_embedding_max_norm = 1
    entity_embedding_norm_type = 2
    relation_embedding_max_norm = 1
    relation_embedding_norm_type = 2
    hyper_params = BaseModule.hyper_params + [RELATION_EMBEDDING_DIM, SCORING_FUNCTION_NORM]

    def __init__(self, config: Dict) -> None:
        super().__init__(config)
        config = TransDConfig.from_dict(config)

        # Embeddings
        self.relation_embedding_dim = config.relation_embedding_dim

        # max_norm = 1 according to the paper
        self.relation_embeddings = nn.Embedding(self.num_relations, self.relation_embedding_dim,
                                                norm_type=self.relation_embedding_norm_type,
                                                max_norm=self.relation_embedding_max_norm)
        self.projection_matrix_embs = nn.Embedding(self.num_relations, self.relation_embedding_dim * self.embedding_dim)
        self.scoring_fct_norm = config.scoring_function_norm

        self._initialize()

    def _initialize(self):
        entity_embeddings_init_bound = relation_embeddings_init_bound = 6 / np.sqrt(self.embedding_dim)
        nn.init.uniform_(
            self.entity_embeddings.weight.data,
            a=-entity_embeddings_init_bound,
            b=entity_embeddings_init_bound,
        )
        nn.init.uniform_(
            self.relation_embeddings.weight.data,
            a=-relation_embeddings_init_bound,
            b=relation_embeddings_init_bound,
        )

        norms = torch.norm(self.relation_embeddings.weight, p=2, dim=1).data
        self.relation_embeddings.weight.data = self.relation_embeddings.weight.data.div(
            norms.view(self.num_relations, 1).expand_as(self.relation_embeddings.weight))

    def _compute_scores(self, h_embs, r_embs, t_embs):
        # Add the vector element wise
        sum_res = h_embs + r_embs - t_embs
        distances = torch.norm(sum_res, dim=1, p=self.scoring_fct_norm).view(size=(-1,))
        distances = torch.mul(distances, distances)
        return distances

    def _compute_loss(self, pos_scores, neg_scores):
        # y == -1 indicates that second input to criterion should get a larger loss
        # y = torch.Tensor([-1]).cuda()
        # NOTE: y = 1 is important
        # y = torch.tensor([-1], dtype=torch.float, device=self.device)
        y = np.repeat([-1], repeats=pos_scores.shape[0])
        y = torch.tensor(y, dtype=torch.float, device=self.device)

        # Scores for the psotive and negative triples
        pos_scores = torch.tensor(pos_scores, dtype=torch.float, device=self.device)
        neg_scores = torch.tensor(neg_scores, dtype=torch.float, device=self.device)
        # neg_scores_temp = 1 * torch.tensor(neg_scores, dtype=torch.float, device=self.device)

        loss = self.criterion(pos_scores, neg_scores, y)
        return loss

    def _project_entities(self, entity_embs, projection_matrix_embs):
        projected_entity_embs = torch.einsum('nk,nkd->nd', [entity_embs, projection_matrix_embs])
        projected_entity_embs = torch.clamp(projected_entity_embs, max=1.)
        return projected_entity_embs

    def predict(self, triples):
        # triples = torch.tensor(triples, dtype=torch.long, device=self.device)
        heads = triples[:, 0:1]
        relations = triples[:, 1:2]
        tails = triples[:, 2:3]

        head_embs = self.entity_embeddings(heads).view(-1, self.embedding_dim)
        relation_embs = self.relation_embeddings(relations).view(-1, self.relation_embedding_dim)
        tail_embs = self.entity_embeddings(tails).view(-1, self.embedding_dim)

        proj_matrix_embs = self.projection_matrix_embs(relations).view(-1, self.embedding_dim,
                                                                       self.relation_embedding_dim)

        proj_heads_embs = self._project_entities(head_embs, proj_matrix_embs)
        proj_tails_embs = self._project_entities(tail_embs, proj_matrix_embs)

        scores = self._compute_scores(h_embs=proj_heads_embs, r_embs=relation_embs, t_embs=proj_tails_embs)

        return scores.detach().cpu().numpy()

    def forward(self, batch_positives, batch_negatives):
        pos_heads = batch_positives[:, 0:1]
        pos_relations = batch_positives[:, 1:2]
        pos_tails = batch_positives[:, 2:3]

        neg_heads = batch_negatives[:, 0:1]
        neg_relations = batch_negatives[:, 1:2]
        neg_tails = batch_negatives[:, 2:3]

        pos_h_embs = self.entity_embeddings(pos_heads).view(-1, self.embedding_dim)
        pos_r_embs = self.relation_embeddings(pos_relations).view(-1, self.relation_embedding_dim)
        pos_t_embs = self.entity_embeddings(pos_tails).view(-1, self.embedding_dim)

        neg_h_embs = self.entity_embeddings(neg_heads).view(-1, self.embedding_dim)
        neg_r_embs = self.relation_embeddings(neg_relations).view(-1, self.relation_embedding_dim)
        neg_t_embs = self.entity_embeddings(neg_tails).view(-1, self.embedding_dim)

        proj_matrix_embs = self.projection_matrix_embs(pos_relations).view(-1, self.embedding_dim,
                                                                           self.relation_embedding_dim)

        # Project entities into relation space
        proj_pos_heads_embs = self._project_entities(pos_h_embs, proj_matrix_embs)
        proj_pos_tails_embs = self._project_entities(pos_t_embs, proj_matrix_embs)

        proj_neg_heads_embs = self._project_entities(neg_h_embs, proj_matrix_embs)
        proj_neg_tails_embs = self._project_entities(neg_t_embs, proj_matrix_embs)

        pos_scores = self._compute_scores(h_embs=proj_pos_heads_embs, r_embs=pos_r_embs, t_embs=proj_pos_tails_embs)
        neg_scores = self._compute_scores(h_embs=proj_neg_heads_embs, r_embs=neg_r_embs, t_embs=proj_neg_tails_embs)

        loss = self._compute_loss(pos_scores=pos_scores, neg_scores=neg_scores)
        return loss



# -*- coding: utf-8 -*-

"""Utilities for predicting links."""

from itertools import product

import numpy as np
import pandas as pd
import torch

from pykeen.utilities.pipeline import load_data
from pykeen.utilities.triples_creation_utils import create_mapped_triples

__all__ = [
    'create_triples',
    'make_predictions',
]


def create_triples(entity_pairs, relation) -> np.ndarray:
    subjects = entity_pairs[:, 0:1]
    objects = entity_pairs[:, 1:2]
    relation_repeated = np.reshape(np.repeat(relation, repeats=subjects.shape[0]), newshape=(-1, 1))
    return np.concatenate([subjects, relation_repeated, objects], axis=1)


def make_predictions(
        kge_model,
        entities,
        relations,
        entity_to_id,
        rel_to_id, device,
        remove_training_triples=False,
        training_set_path=None,
):
    all_entity_pairs = np.array(list(product(entities, entities)))

    if relations.size == 1:
        all_triples = create_triples(entity_pairs=all_entity_pairs, relation=relations)
    else:
        all_triples = create_triples(entity_pairs=all_entity_pairs, relation=relations[0])

        for relation in relations[1:]:
            triples = create_triples(entity_pairs=all_entity_pairs, relation=relation)
            all_triples = np.append(all_triples, triples, axis=0)

    if remove_training_triples:
        assert (training_set_path is not None)
        training_triples = load_data(training_set_path)
        training_triples = pd.DataFrame(data=training_triples)
        all_triples = pd.DataFrame(data=all_triples)
        merged = all_triples.merge(training_triples, indicator=True, how='outer')
        merged = merged[merged['_merge'] == 'left_only'].values
        all_triples = np.array(merged[:, :-1], dtype=np.str)

    mapped_triples, _, _ = create_mapped_triples(all_triples, entity_label_to_id=entity_to_id, relation_label_to_id=rel_to_id)

    mapped_triples = torch.tensor(mapped_triples, dtype=torch.long, device=device)

    id_to_entity = {value: key for key, value in entity_to_id.items()}
    id_to_relation = {value: key for key, value in rel_to_id.items()}
    subject_column = np.vectorize(id_to_entity.get)(mapped_triples[:, 0:1])
    predicate_column = np.vectorize(id_to_relation.get)(mapped_triples[:, 1:2])
    object_column = np.vectorize(id_to_entity.get)(mapped_triples[:, 2:3])

    triples_ordered_by_ids = np.concatenate([subject_column, predicate_column, object_column], axis=1)

    predicted_scores = kge_model.predict(mapped_triples)

    _, sorted_indices = torch.sort(torch.tensor(predicted_scores, dtype=torch.float),
                                   descending=False)

    sorted_indices = sorted_indices.cpu().numpy()

    ranked_triples = triples_ordered_by_ids[sorted_indices, :]

    subs = np.reshape(ranked_triples[:, 0:1], newshape=(-1))
    objs = np.reshape(ranked_triples[:, 2:3], newshape=(-1))

    indices = np.where(subs != objs)[0]

    ranked_scores = np.reshape(predicted_scores[sorted_indices], newshape=(-1, 1))
    ranked_triples = np.concatenate([ranked_triples, ranked_scores], axis=1)

    return ranked_triples[indices, :]
